#include <zlib.h>
// #include "mm2-fast/minimap.h"
// #include "mm2-fast/bseq.h"
#include "mm2-fast/kdq.h"
#include "mm2-fast/khash.h"
#include "mm2-fast/kseq.h"
#include "mm2-fast/ksw2.h"
#include "mm2-fast/kvec.h"
#include "mm2-fast/mmpriv.h"
#include "mm2-fast/kalloc.h"
#include "mm2-fast/ketopt.h"
#include "mm2-fast/krmq.h"
#include "mm2-fast/ksort.h"
#include "mm2-fast/kthread.h"
#include "mm2-fast/sdust.h"
