#include <zlib.h>
#include "minimap2/minimap.h"
#include "minimap2/bseq.h"
#include "minimap2/kdq.h"
#include "minimap2/khash.h"
#include "minimap2/kseq.h"
#include "minimap2/ksw2.h"
#include "minimap2/kvec.h"
#include "minimap2/mmpriv.h"
#include "minimap2/kalloc.h"
#include "minimap2/ketopt.h"
#include "minimap2/krmq.h"
#include "minimap2/ksort.h"
#include "minimap2/kthread.h"
#include "minimap2/sdust.h"
